#!/usr/bin/perl
#
#    FEARLUS Expt Type 1 v. 0-5-1-5: run.pl
#    Copyright (C) 2007  Macaulay Institute
#
#    This file is part of FEARLUS Expt Type 1 v. 0-5-1-5, experiment scripts
#    for multiple runs with FEARLUS.
#
#    FEARLUS Expt Type 1 v. 0-5-1-5 is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    FEARLUS Expt Type 1 v. 0-5-1-5 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details. (LICENCE file in
#    this directory.)
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Contact information:
#      Gary Polhill,
#      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
#      g.polhill@macaulay.ac.uk
#
#
# run.pl
#
# Gary Polhill 15 June 2001
#
# This is a script to do runs pitting one subpopulation against another in a
# number of different environments. It expects the corresponding create.pl to
# have been run already in the current working directory.
#
# The program's usage is as follows:
#
# run.pl [-history[=<number of history runs to save>]] [-restart]
#        [-command=<path to FEARLUS command>]
#        [-onlydirs=<list of sp directories to run>]
#        <number of runs to do>

use strict;
use Cwd;
require "tools.pl";

srand(time);
# Global variable declarations

# $history is -1 for all runs, 0 for no runs, +ve for first n runs
my $history = 0;

# $restart is 0 to continue with runs so far and 1 to start from the beginning
my $restart = 0;

# $exe is the command to use to run the FEARLUS model
my $exe = '/home/gp/swarm/fearlus/model0-5-1-5/model0-5-1-5';

# Number of runs
my $runs;

# SP directories to do
my @onlyspdirs;
my %runspdirs;
my $allspdirs = 1;

# Check the arguments to the program

my $arg;

$arg = shift(@ARGV) || &usage("You must supply the number of runs");
while($arg =~ /^-/) {
  if($arg =~ /^-history/) {
    if($arg =~ /^-history=([0-9]+)$/) {
      $history = $1;
    }
    elsif($arg =~ /^-history.+/) {
      &usage("Garbled history argument: $arg");
    }
    else {
      $history = -1;
    }
  }
  elsif($arg =~ /^-onlydirs=(.+)$/) {
    my $dirlist = $1;
    @onlyspdirs = split(/,/, $dirlist);
    $allspdirs = 0;
    $restart = 1;
    foreach my $dir (@onlyspdirs) {
      if(!-d "$dir") {
	die "$dir is not a valid sub-population directory\n";
      }
      $runspdirs{$dir} = $dir;
    }
  }
  elsif($arg eq '-restart') {
    $restart = 1;
  }
  elsif($arg =~ /^-command=(.+)$/) {
    $exe = $1;
  }
  $arg = shift(@ARGV) || &usage("You must supply the number of runs");
}

if($arg =~ /^[0-9]+$/) {
  $runs = $arg;
}
else {
  &usage("Invalid number of runs: $arg");
}

# Check the command is runnable

if(!-x "$exe") {
  die "$exe is not an existing file that you have execute permissions for".
    ", so I cannot run the model\n";
}
if($exe !~ /\//) {
  die "$exe must be an absolute path to the FEARLUS model executable\n";
}

# Get the list of environments we are comparing the two subpopulations in

my @envs = glob("*.model");
my @shortenvs = &get_short_envs(@envs);

# Create a log file to save messages in. If the log file exists, check progress
# through the experiment. Log file entries indicating successful completion
# of a run have the format <DATE> Completed run <RUN> of environment <ENV> in
# directory <DIR>
#
# The presence of a log file will be used to allow continuation of interrupted
# experiments

# $oktogo is a flag that tells the loop it is ready to run. It will only be 0
# at the beginning of the loop if there are a number of runs that have been
# done already.
my $oktogo = 1;
my($godate, $gorun, $goenv, $godir);
if(-f "run.log" && !$restart) {
  open(LOG, "<run.log") || die "Cannot read log file run.log: $!\n";
  while(my $line = <LOG>) {
    if($line 
       =~ m/^(.+) Completed run (.+) of environment (.+) in directory (.+)$/) {
      # We just want the last completed run that appears in the file
      ($godate, $gorun, $goenv, $godir) = ($1, $2, $3, $4);
      $oktogo = 0;
    }
  }
  close(LOG);
  open(LOG, ">>run.log") || die "Cannot append to log file run.log: $!\n";
  print LOG &datetime()." Restarting after run $gorun of environment $goenv".
    " in directory $godir\n";
}
elsif(-f "run.log") {
  open(LOG, ">>run.log") || die "Cannot append to log file run.log: $!\n";
}
else {
  open(LOG, ">run.log") || die "Cannot create log file run.log: $!\n";
}

if(!$allspdirs) {
  print LOG &datetime()." Running experiments in the following subpopulation ".
    "directories only: ".join(", ", sort(keys(%runspdirs)))."\n";
}

my $oldfh = select LOG;
# Set autoflush on LOG to 1 -- this will flush the buffer after every print.
$| = 1;
select $oldfh;

# Loop through the directories and run the experiments therein

foreach my $spdir (sort(glob("*-v-*"))) {
  # The create script skips over matching files that aren't directories
  next if !-d "$spdir";
  next if !($allspdirs || defined($runspdirs{$spdir}));
  # Loop through all the environments
  for(my $env = 0; $env <= $#shortenvs; $env++) {
    my $envdir = "$shortenvs[$env]";
    # Check the environment contest sub-directory exists
    if(!-d "$spdir/$envdir") {
      die "Cannot find directory $spdir/$envdir -- have you run the ".
	"creation script (cwd = ".getcwd().")?\n";
    }
    # Loop again if we're not OK to go
    next if (!$oktogo && $godir ne "$spdir/$envdir");
    # Change into the experiment contest sub-directory
    chdir "$spdir/$envdir" 
      || die "Cannot change directory to $spdir/$envdir: $!\n";
    # Loop through all the runs
    for(my $run = 1; $run <= $runs; $run++) {
      next if (!$oktogo && $gorun != $run);
      next if (!$oktogo && $goenv ne $envs[$env]);
      # If we get here and we're not $oktogo, then we're now at the point
      # of the last successful run. Set $oktogo to 1 and loop again.
      if(!$oktogo && $goenv eq $envs[$env]) {
	print LOG &datetime()." Reached run $run of environment $envs[$env] ".
	  "in directory $spdir/$envdir\n";
	$oktogo = 1;
	next;
      }
      # If we get here, then we're $oktogo. Get the modelfile name (it's just
      # the name of the environment)
      my $modelfile = $envs[$env];
      # Set up the other commandline argument filenames
      my $reportfile = "$spdir-$shortenvs[$env]-run$run-rep";
      my $historyfile = "$spdir-$shortenvs[$env]-run$run";
      # Delete any files that the report or history files might over-write
      if(-e "$reportfile.txt") {
	unlink "$reportfile.txt"
	  || die "Cannot remove $reportfile.txt in directory ".
	    "$spdir/$envdir: $!\n";
      }
      # Delete the history file regardless of whether it exists from an
      # earlier run -- it won't correspond to this run, and that could be
      # confusing
      if(-e "$historyfile.txt") {
	unlink "$historyfile.txt"
	  || die "Cannot remove $historyfile.txt in directory ".
	    "$spdir/$envdir: $!\n";
      }
      # Create a temporary file to store any error messages
      my $tmpfile;
      do {
	$tmpfile = sprintf("/tmp/FEARLUS%06d", int(rand() * 999999));
      } while(-e "$tmpfile");
      # Set up the command to run
      my $command = "$exe --batch --withseed=TIME ".
	"--repconfig=Parcel.repcfg --report=$reportfile ".
	  "--parameters=$modelfile ".
	    (($history < 0) || ($history >= $run) ?
	     "--history=$historyfile " : "").
	       "> /dev/null 2> $tmpfile";
      print LOG &datetime()." Running command $command\n";
      # Run the command
      my $retval = system("$command");
      if($retval != 0) {
	print LOG &datetime()." Command failed with exit code $retval:\n";
	open(TMP, "<$tmpfile") 
	  || die "Cannot read tempfile $tmpfile: $!\n";
	while(<TMP>) {
	  print LOG &datetime()."   $_";
	}
	close(TMP);
      }
      else {
	print LOG &datetime()." Completed run $run of environment ".
	  "$envs[$env] in directory $spdir/$envdir\n";
      }
      unlink($tmpfile) || die "Cannot delete tempfile $tmpfile: $!\n";
    }
    chdir "$ENV{PWD}" 
      || die "Cannot change directory back to $ENV{PWD}: $!\n";
  }
}

exit 0;

sub usage {
  my($message) = @_;
  die <<DEATH;
$message

This script is called as follows:
$0 [-history[=<number of history runs to save>]] [-restart] 
   [-command=<full path to FEARLUS model executable>]
   [-onlydirs=<list of sp directories to run>] <number of runs to do>

-history is an optional argument that will tell FEARLUS to save a history file
for each run that it does. You can optionally specify that the first NN runs
save a history (whilst the rest do not) by giving -history=NN. Be warned that
history files can occupy a lot of disk space, especially when making a number
of runs.

-restart will cause the script to restart the experiment even if it is part of
the way through it. If you do not give -restart, then the script will check
run.log for successfully completed runs, and not do them again.

-onlydirs allows you to specify a comma separated (no spaces) list of subpop
directories in which to run experiments. Experiments in these directories will
be run regardless of what was done before.

-command specifies an alternative path to the FEARLUS model. The default is
currently $exe.

The final argument is the number of runs to do in each experiment.
DEATH
}
